/**
 * MPP.c
 * Belongs to the TARGETING software
 * http://www.bioinf.uni-leipzig.de/Software/Targeting
 *
 * @author Christian Arnold
 * @email achristian@bioinf.uni-leipzig.de
 * @company Professur fuer Bioinformatik, University of Leipzig
 */
 
   
/*           TARGETING Copyright (C) 2010  Christian Arnold

* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.

* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.

* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>
* or the file "GNU Public License.txt" in the "src" folder.
*/


#include "tree.h"
#include "MPP.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define FILENAME_MWM_IN "MWM.in"
#define FILENAME_MWM_OUT "MWM.out"
#define BUFSIZE_LINE 10
#define ROOT_ID 0
#define VERBOSE 0
#define USE_APPR 1



void fillNMatrix (Node ***N, int numberLeaves, Node **pp_leafList, int root_id) {
		
	for (int i = 0; i < numberLeaves; i++) {
	    //printf("Fill matrix for tip with ID %d...\n",pp_leafList[i]->id);
	    int tip_id = pp_leafList[i]->idL;
	    Node *p_node = pp_leafList[i];
	    
	    Node *parent = pp_leafList[i]->p_parent;
	      
	    while (p_node->idA != root_id) {
		
			//printf("Node ID:%d, Root ID: %d\n",id,root_id);		
			if (!parent) {
			  fprintf(stderr, "Error: Node is NULL\n");
			  exit(EXIT_FAILURE);
			}
			  
			if (p_node->idL != tip_id) {
			  parent  = parent->p_parent;
			}

			//printf("N[%d][%d]=%d\n",id_parent,tip_id,id);
			N[parent->idA][tip_id] = p_node;

			p_node = parent;
	  }	    
	}
}


void calcOptPathSystem (Node *p_node, Node ***F, Node *****F2, Node ***N, double **Q, double ***Q2, double **R, double *S, double **W, int ***MWM, int prec) {

	int idA = p_node->idA;
	int idI = p_node->idI;	
	int deg = p_node->deg;
	
	//printf("Calculating optimal path-system for node with ID %d...\n",idA);
	
	if (F==NULL || N==NULL || R==NULL || S==NULL || W==NULL ) {
	   fprintf(stderr,"Error: At least one matrix equals NULL\n");
	   exit(EXIT_FAILURE); 
	}
	
	//Leaf node or already calculated
	if (deg == 0 || S[idA] >= 0) {
		return;
	}
	
	if (deg == 1 || deg < 0) { //undefined
		
	   fprintf(stderr,"Error: Degree of p_node with ID %d not valid:%d, exit\n",idA,deg);
	   exit(EXIT_FAILURE);	
	
	} else if (deg == 2) { //binary case
		  
		  double S1 = 0;

		  for (int i=0; i< deg;i++) {
			  double score_child = S[p_node->p_children[i]->idA];
			  if (score_child != -1) {
			       S1 += score_child;	       
			  } else {
			       fprintf(stderr,"Error: Score of node %f not yet calculated...\n",score_child);
			       exit(EXIT_FAILURE);
			  }
		  }
 
		  double S2 = 0;		  
		  int counterLeaves = 0;
		  int *p_counterLeaves = &counterLeaves;
		  
		  int leavesInX = 0;
		  int leavesInY = 0;

		  Node **pp_leafList1 = NULL;
		  Node **pp_leafList2 = NULL;

		  pp_leafList1 = getLeavesInSubtree (p_node->p_children[0], pp_leafList1, p_counterLeaves);
		  leavesInX = counterLeaves;
		  counterLeaves = 0;
		  pp_leafList2 = getLeavesInSubtree (p_node->p_children[1], pp_leafList2, p_counterLeaves);
		  leavesInY = counterLeaves;

		  if (pp_leafList1 == NULL || pp_leafList2 == NULL) {
			fprintf(stderr,"Error: NULL exit\n");
			exit(EXIT_FAILURE);	
		  }
		  	  
		  Node *x,*y;
		  int x_idL,y_idL;
		  Node *p_leaf_x = NULL;
		  Node *p_leaf_y = NULL;
		  
		  for (int i = 0; i < leavesInX; i++) {
		      
		       x    = pp_leafList1[i];
		       x_idL = pp_leafList1[i]->idL;
		       
		       for (int j = 0; j < leavesInY; j++) {
				 				 
			      //printf("Loop i=%d j=%d\n",i,j);			    
			      y     = pp_leafList2[j];
			      y_idL = pp_leafList2[j]->idL;
			 
			      double R1 = R[idA][x_idL];
			      if (R1 == -1) {
					R[idA][x_idL] = R1 = R_rek(p_node->p_children[0], x, N, Q, R, S);
			      }
			      
			      double R2 = R[idA][y_idL];
			      if (R2 == -1) {
					R[idA][y_idL] = R2 = R_rek(p_node->p_children[1], y, N, Q, R, S);
			      }
			      
			      //printf("W[%d][%d]:%f, R1=%f. R2=%f\n",x_idL, y_idL, W[x_idL][y_idL], R1, R2);
			      double score  = W[x_idL][y_idL] + R1 + R2; 
			      
			      if (score > S2) { //new maximum
					S2 = score;
					p_leaf_x = x;
					p_leaf_y = y;
			      }
			 }
		  }

		  free(pp_leafList1);
		  free(pp_leafList2);
		  
		  //printf("Score S1: %f, S2:%f\n",S1,S2);
		 		    
		  if (S2 > S1) {
			F[idI][0] = p_leaf_x;
			F[idI][1] = p_leaf_y;
		  } //else : already set to NULL in initialization step

		  S[idA] = max(S1,S2);
		  
		  //printf("Score of optimal path-system for node with ID %d: %f\n",idA,S[idA]);
	
	} else if (deg > 2) { //multifurcation
	     
	       //fprintf(stderr,"Error: Node has more than two children, not implemented yet\n");
		   //printf(" Node with ID %d has more than two children...\n",idA);
	       //exit(EXIT_FAILURE);
		   
		   //S[idA] = 1;
		  // return;
	       
	       if (F2 == NULL || Q2 == NULL || S == NULL || MWM == NULL) {
				fprintf(stderr,"Error: Matrix NULL\n");
				exit(EXIT_FAILURE); 
	       }
		
			//Calculate MWM for node without excluding any children, this corresponds to calculating S_u
			S[p_node->idA] = calcMWM(p_node, NULL, F2, N, Q, Q2, R, S, W, MWM, prec);
			
			//Calculate MWM for node with excluding one child k, this corresponds to calculating Q_ux
   			for (int i = 0; i < deg; i++) {			
				double scoreMWM = calcMWM(p_node, p_node->p_children[i], F2, N, Q, Q2, R, S, W, MWM, prec); 

				//determine all leaves x \in T[u_x] and set Q_{u,x}= MWM(chd(u)\{u_x})
				
				int numberLeavesInSubtree  = 0;
				int *p_counterLeaves = &numberLeavesInSubtree;
				Node **pp_leafList = getNextLeafRek(p_node->p_children[i],NULL,p_counterLeaves, NULL);
				
				for (int j = 0; j < numberLeavesInSubtree; j++) {
					Q[p_node->idA][pp_leafList[j]->idL] = scoreMWM;				
				}
				
				free(pp_leafList);
					
			}
	}

}


double R_rek (Node *p_node,Node *p_targetLeaf, Node ***N, double **Q, double **R, double *S) {
    
	//printf("R_rek: Current ID:%d )idM:%d) deg:%d Leaf:%d\n",p_node->idA, p_node->idM, p_node->deg,p_targetLeaf->idA );
  
	if (p_node->idA != p_targetLeaf->idA) { //target leaf is not yet reached
	  
		  Node *p_nextNodeBar = NULL;
		  int idA_child = 0;
		  double score = 0;
		  Node *p_nextNode = N[p_node->idA][p_targetLeaf->idL];
		  
		  if (p_nextNode == NULL) {
			  fprintf(stderr,"Error:  NULL pointer in p_nextNode\n");
			  exit(EXIT_FAILURE);        	       
		  }
	    
	      if (p_node->deg == 2) { //binary case: R_ux = R_{u_xx} + S_{u_{x\bar}}
			  p_nextNodeBar = getUquer(p_node, p_targetLeaf, N);
			  idA_child = p_nextNodeBar->idA;
			  
			  if ( p_nextNodeBar == NULL) {
				   fprintf(stderr,"Error: NULL pointer in p_nextNodeBar or p_nextNode\n");
				   exit(EXIT_FAILURE);        	       
			  }
			  if (S[idA_child] == -1) {
			   fprintf(stderr,"Error: S not calculated yet for node with ID %d and degree %d\n", p_node->idA, p_node->deg);
			   exit(EXIT_FAILURE);        	       
		      }
			  
			  score = S[idA_child];
			  
		  } else { //multifurcating case: R_ux = R_{u_xx} + Q_{ux}
			  
			  if ( Q[p_node->idA][p_targetLeaf->idL] == -1) {
				   fprintf(stderr,"Error: NULL pointer in Q[%d][%d]: %lf\n",p_node->idA, p_targetLeaf->idL, Q[p_node->idA][p_targetLeaf->idL]);
				   exit(EXIT_FAILURE);    
		      }
			  
			  score = Q[p_node->idA][p_targetLeaf->idL];
		  }

		  R[p_node->idA][p_targetLeaf->idL] = R_rek(p_nextNode,p_targetLeaf,N,Q,R,S) + score;

		  return R[p_node->idA][p_targetLeaf->idL];	  
	 } 
	 else {
		return 0;
	 }
	   
}

     
double calcMWM (Node *p_node, Node *p_nodeExcluded, Node *****F2, Node ***N, double **Q, double ***Q2, double **R, double *S, double **W, int ***MWM, int prec) {
	
	if (VERBOSE)
		;//printf("Calculating MWM for node with ID %d and degree %d,excluding node %d\n",p_node->idA, p_node->deg, p_nodeExcluded == 0? "(none)": p_nodeExcluded->idA);
	
	int apprWeights = prec == -1? 0 : 1;
		
	int deg = p_nodeExcluded == NULL? (p_node->deg) : (p_node->deg-1);
	int idM = p_node->idM;
	int n_vertices = deg * 2;
	int n_edges = (deg * (deg - 1))/2 + deg;
	int n_deg1 = deg, n_deg2 = 1;
	
	int MWMIndex = getMWMIndex(p_node, p_nodeExcluded);
	
	/*
	 * ----------------------------------
     * 1. Compute edge weights for MWM
	 * ----------------------------------
	 */
	
	//Create list of leaves
	Node ***p_leafList = malloc (deg * sizeof(Node **));
	int arr_sizes[deg];
	int counterLeaves = 0;
	int *p_counterLeaves =  &counterLeaves;
	for (int i = 0; i < deg; i++) {  
	  counterLeaves = 0;
	  p_leafList[i] = getLeavesInSubtree (p_node->p_children[i], NULL, p_counterLeaves);
	  arr_sizes[i] = counterLeaves;
	  if (arr_sizes[i] == 0 || p_leafList[i] == NULL) {
		fprintf(stderr,"  \nError: No leaf in subtree with root %d, exit...\n",p_node->p_children[i]->idA);
	    exit(EXIT_FAILURE);		
	  }
	}
	
	Node *x,*y;
	int x_idL,y_idL;
	int leavesInX = 0;
	int leavesInY = 0;
	double curMax = 0;
	int foundBetterMax = 0;
	
	//For all pairs of children v, v' determine the pair of leaves (k,l) that maximizes Q
	
	for (int i = 0; i < deg; i++) { //for all children v
  
	  if (p_node->p_children[i] == p_nodeExcluded) { //skip node
		continue;
	  }
				  
	  leavesInX = arr_sizes[i];
	  Node *p_i = p_node->p_children[i];
	  int i_idA = p_i->idA;
	  
	  for (int j = i + 1; j < deg; j++) { //for all children v'
  
		  foundBetterMax = 0;
		  
		  if (p_node->p_children[j] == p_nodeExcluded) { //skip node
		  continue;
		  }  
		  
		  leavesInY = arr_sizes[j];
		  Node *p_j = p_node->p_children[j];
		  int j_idA = p_j->idA;

		  //Determine pair of leaves that maximizes Q_{v,v'} with v=i and v'=j
		  curMax = 0;
		  Q2[idM][i][j] = Q2[idM][j][i] = curMax;
		  //printf("Init: Q2[%d][%d][%d] = Q2[%d][%d][%d] = 0\n",idM,i,j, idM,j,i);
		  
		  //printf("Determine best scoring pair in children %d and %d\n", p_i->idA, p_j->idA);
		  
		  foundBetterMax = 0;
		  
		  for (int k = 0; k < leavesInX; k++) { //all leaves k in T[v]
		      
			   x     = p_leafList[i][k];
		       x_idL = p_leafList[i][k]->idL;
		       
		       for (int l = 0; l < leavesInY; l++) { //all leaves l in T[v']
				 				 
			      //printf("LeavesInX: %d, LeavesInY:%d Loop i=%d j=%d k=%d l=%d\n",leavesInX, leavesInY, i,j,k ,l);
			    
			      y     = p_leafList[j][l];
			      y_idL = p_leafList[j][l]->idL;
				  
				  //printf("ID x:%d, ID y:%d\n",x->idA, y->idA);
			 
			      double R1 = R[i_idA][x_idL];
			      if (R1 == -1) {
					R[i_idA][x_idL] = R1 = R_rek(p_i, x, N, Q, R, S);
			      }
			      
			      double R2 = R[j_idA][y_idL];
			      if (R2 == -1) {
					R[j_idA][y_idL] = R2 = R_rek(p_j, y, N, Q, R, S);
			      }
			      
			      //printf("W[%d][%d]:%f, R1=%f. R2=%f\n",x_idL, y_idL, W[x_idL][y_idL], R1, R2);
			      double score  = W[x_idL][y_idL] + R1 + R2; 
				  
				  //printf("Score for children pair (%d,%d): R1=%lf, R2=%lf, weight=%lf\n", k, l, R1, R2, W[x_idL][y_idL]);
			      
			      if (score >= curMax) { //new maximum
					Q2[idM][i][j] = Q2[idM][j][i] = curMax = score;
					//printf("Set Q2[%d][%d][%d]= %lf\n", idM, i, j, score);
					F2[idM][i][j][0] = F2[idM][j][i][0] = x;
					F2[idM][i][j][1] = F2[idM][j][i][1] = y;
					foundBetterMax = 1;
			      }
			 
			   } // end l loop
		  
		  } // end k loop
		  
		  if (foundBetterMax == 0) {
			  fprintf(stderr,"  \nError: Better maximum not found for i=%d and j=%d, exit...\n",i, j);
			  exit(EXIT_FAILURE);						
		  }
		  
	  } // end v' loop
	    
	} // end v loop
	

    // Free memory
	for (int i = 0; i < deg; i++) {  
	  free(p_leafList[i]);	  
	}
	free(p_leafList);

	
	/*
	 * ----------------------------------
	 * 2. Write input file for MWM solver
	 * ----------------------------------
	 */
	
	/*
	 *
	 * Input format
	 *	
	 *There are three types of lines: the first line, vertex lines, 
	 *and edge lines. The fields in each line type are as follows. 

	 *First line-> size edges U
		*size: integer giving number of vertices
		*edges: integer giving number of edges 
		*U: character ``U'' or ``u'' specifying an undirected graph

	 *Vertex lines->  degree vlabel xcoord ycoord
		*degree: edge degree of the vertex
		*vlabel: vertex label (ignored--vertices are referred to by index)
		*xcoord: integer x-coordinate location (ignored)
		*ycoord: integer y-coordinate location (ignored) 

		*Each vertex line is followed immediately by the lines 
		*for all its adjacent edges (thus each edge appears twice, 
		*once for each vertex).

	 *Edge lines-> adjacent  weight
		*adjacent: index (not vlabel) of the adjacent vertex
		*weight: integer edge weight
	
	 *Example:
		6 8 U
		2 3 0 0 
		2 6
		3 8
		3 3 0 0 
		1 6 
		5 3
		4 6
		3 3 0 0 
		1 8
		4 3
		5 3
		3 3 0 0 
		2 6
		3 3
		6 8
		3 3 0 0 
		3 3
		2 3
		6 6
		2 3 0 0 
		4 8
		5 6
		
	 *Convention in this program:
		*Consider a multifurcating node u with k children that are included in the matching problem
		*The first k nodesp_nodeList[j]->idA of the MWM correspond to the real children of u, with IDs 1-k and a degree of k each (k-1 to the other children + 1 for the (v,v*) edge)
		*Nodes (k+1)-2k correspond to the auxiliary nodes v*, with ID (k+1)-2k and degree 1		
	 */
	
	remove(FILENAME_MWM_IN);
	FILE *p_File = fopen (FILENAME_MWM_IN, "a+");
	if (!p_File) {
		fprintf(stderr,"  \nError: Could not open output file \"%s\", exit...\n",FILENAME_MWM_IN);
	    exit(EXIT_FAILURE);
	}
	
	//First line
	fprintf (p_File,"%d %d U\n", n_vertices, n_edges);
	if (VERBOSE) {
	  printf ("   %d %d U\n", n_vertices, n_edges);
	}

	//Vertex lines	
	for (int i = 0; i < n_vertices; i++) {
		
		int excludedNode = 0;
		
		//Get child node that corresponds to i
		int index = i < deg? i : (i-deg);
		
		if (p_node->p_children[index] == p_nodeExcluded) {
		  excludedNode = 1;
		}
	
		int degNode = i < deg? n_deg1: n_deg2; //Real nodes v and auxiliary nodes v* have a different degree
		
		fprintf (p_File,"%d 0 0 0\n", degNode);
		if (VERBOSE) {
		  printf ("   %d 0 0 0\n", degNode);
		}
		
		degNode = i < deg? (degNode + 1): degNode; // increase degNode to 1 for the next loop (due to the extra edge to v* and because edge must be to a different node)

		//Edge lines for vertices v
		for (int j = 0; j < degNode; j++) {

			if (i == j) { //edge must go to a different node
			 continue;
			}

			int adjVertex = -1;

			// Two cases have to be distinguished
			if (i < deg) { //1. Vertices of type v
			 adjVertex = j < deg ? (j + 1) : ((i + 1) + deg); 
			} else { //1. Vertices of type v'
			 adjVertex = (i + 1) - deg;
			}

			double weight;
			int index1 = i, index2 = j;

			if (excludedNode == 1) {
			  //printf("Increase indexes by 1: i=%d j=%d\n",i,j);
			  index1++;
			  index2++;
			}
		   
			if (i < deg) { //normal node v
			  
			  if (j < deg) { // edge (v,v')
				//printf("Take value Q2[%d][%d][%d]= %lf\n",idM,i,j,Q2[idM][i][j]);
				if (Q2[idM][index1][index2] == -1) {
					  fprintf(stderr,"Error: Q2 equals -1, exit...\n");
					  exit(EXIT_FAILURE);
				  }
				  weight = Q2[idM][index1][index2]; // Edges for (v,v')	
				  if (apprWeights) {
					weight = ceil (prec * weight);
				  }
				  if (weight == 0) {
					fprintf(stderr,"  \nError: Zero weight for (v,v') edge: Q2[%d][%d][%d]: %lf, exit...\n",idM, index1, index2, Q2[idM][index1][index2]);
					exit(EXIT_FAILURE);
				  }
			  
			  } else { // edge (v,v*)
				  weight = S[p_node->p_children[index1]->idA];
				  if (apprWeights) {
					weight = ceil (prec * weight);
				  }
				  
			  }
			} 
			else { // node v*
			  weight = S[p_node->p_children[index1-deg]->idA];// Edges for  (v,v*)
			  //Format weight: Take only 2 decimal places after comma
			  if (apprWeights) {
				weight = ceil (prec * weight);
			  }
				
			}
		
		   fprintf (p_File,"%d %lf\n", adjVertex, weight);
		   if (VERBOSE) {
			printf ("   %d %lf\n", adjVertex, weight);
		   }
		}

	}
	
	fclose(p_File);
	
    
	/*
	 * 3. Make system call to MWM solver from Ed Rothberg
	 */
	
	if (VERBOSE)
		printf("Making system call and calculate MWM...\n");
	
	int errorCode = system("./w-match/wmatch MWM.in > MWM.out");
	
	if (errorCode) {
	  ; // prevent compiler warning
	}

	/*
	 * -----------------------------------------
	 * 4. Parse output file and save MWM results
	 * -----------------------------------------
     */
	
	
	/*
	 * The i-th line:  
	 *	Contains the i-th vertex and its mate.
	 *	If the first entry is 0, the i-th vertex is unmatched.
	 *For example:
	 *	1 3
	 *	2 4 
	 *	3 1
	 *	4 2
	 *	5 6
	 *	6 5
	 */
	
	int v1,v2;
	double score = 0;
	int lineCounter = 0, pairCounter = 0;
	char buff[BUFSIZE_LINE];
	p_File = fopen (FILENAME_MWM_OUT, "r");
	if (!p_File) {
		fprintf(stderr,"  \nError: Could not open output file \"%s\", exit...\n",FILENAME_MWM_OUT);
	    exit(EXIT_FAILURE);
	}

	if (VERBOSE) { //Print MWM result on stdout if VERBOSE mode active
	
	  printf("MWM calculated, Result:\n");
	  
	  while (fgets(buff, sizeof (buff), p_File) != NULL ) {

			if ( sscanf( buff, "%d %d", &v1, &v2) == 2 ) {
			  
			  printf("   %d %d\n", v1, v2);
			
			} else {
				fprintf(stderr,"  \nError: Output format not supported (line %s)...\n", buff);
				//exit(EXIT_FAILURE);		
			}			
		}	
	}
	
	//Reset file pointer to beginning of file
	rewind(p_File);
	
	while (fgets(buff, sizeof (buff), p_File) != NULL ) {
	  
		//printf("Scanning next line...\n");

		if ( sscanf( buff, "%d %d", &v1, &v2) == 2 ) {
		  
		  //printf("OK, line parsed...Counter:%d\n",pairCounter);  
		  
		  lineCounter++;		
			
			if (v1 != 0) { //Node is matched
			
				if (v1 > v2) {
				  continue;
				}
										
			    //printf("Parsing line \"%d %d\" from MWM.out...\n", v1, v2);
			  
			  //Decrement IDs because MWM node labels start with 1 and not 0
			    v1--; v2--;
	   			
				if (v2 - v1 == deg) { //Edge (v,v*)
				  
				  //Check if v1 needs to be increased, v2 can be ignored all together
				  if (MWMIndex != 0 && v1 >= (MWMIndex-1)) {
					  //printf("Increase v1...\n");
					  v1++;
				  }
				  
				  //printf("Using S with i=%d  ( edge (v,v*) ), value: %lf\n", p_node->p_children[v1]->idA,S[p_node->p_children[v1]->idA]);
				  score += S[p_node->p_children[v1]->idA];
				  v2 = v1; //internal representation is edge to the node itself
				 
				} 
				else { //Edge (v,v')
				  
				   //increment v2 to match real child ids if a particular node has been excluded in the matching			
				  if (MWMIndex != 0 && v1 >= (MWMIndex-1)) {
					//printf("Increase v1...\n");
					v1++;
				  }
				  
				  if (MWMIndex != 0 && v2 >= (MWMIndex-1)) {
					//printf("Increase v2...\n");
					v2++;
				  }
				  
				  //printf("Using Q2 with i=%d and j=%d ( edge (v,v') ), value: %lf\n", v1, v2, Q2[idM][v1][v2]);
				  score += Q2[idM][v1][v2]; //FALSCH?
				}
				
				
				//printf("!Set MWM[%d][%d][%d]= %d and vica versa\n",idM, MWMIndex, v1, v2);
				//MWM[idM][MWMIndex][pairCounter] = MWM[idM][MWMIndex][v2] = pairCounter;
				MWM[idM][MWMIndex][v1] = v2;
				MWM[idM][MWMIndex][v2] = v1;
							
				pairCounter++;
			
			} 
			else { //Node is unmatched, should not happen in this application
			    //MWM[idM][MWMIndex][pairCounter] = -1;
				fprintf(stderr,"  \nError:Node unmatched in MWM, exit...\n");
				exit(EXIT_FAILURE);	
			  
			}

		} 
		else {
			fprintf(stderr,"  \nError: Wrong output format, exit...\n");
			exit(EXIT_FAILURE);		
		}
	
	}
	
	fclose(p_File);
	
	if (VERBOSE)
	  printf("Score of MWM: %lf\n", score);
	
	if (score == 0) {
		fprintf(stderr,"  \nError: Score of MWM equals 0, exit...\n");
		exit(EXIT_FAILURE);		  
	}
	
	return score;
}

int getMWMIndex (Node *p_node, Node *p_nodeExcluded) {
  
	int MWMIndex = 0;
	if (p_nodeExcluded != NULL) { //MWMIndex different than 0
	  for (int i = 0; i < p_node->deg; i++) {  
		if (p_node->p_children[i] == p_nodeExcluded) {
			MWMIndex = (i + 1);
			break;
		}
		if (i == (p_node->deg - 1)) {
			fprintf(stderr,"  \nError: Could not calculate MWMIndex, exit...\n");
			exit(EXIT_FAILURE);
		}
	  }
	  
	}
	  
	return MWMIndex;

}

void backtrace (Node *p_node, Node *p_nodeExcluded, Node ***F, Node *****F2, Node ***N, double *S, double **W, int ***MWM, char *filename) {
	
			 
     //printf("Backtracing for node %d...\n",p_node->idA);
	 int idI = p_node->idI;
     int idA = p_node->idA;
	 int idM = p_node->idM;
     int deg = p_node->deg;
	 
     FILE *p_File = NULL;
     
     if (idA == ROOT_ID) { //root node
  
		  if (strcmp(filename,"stdout") == 0) {
			   p_File = stdout;	       
		  } 
		  else {
			   p_File = fopen (filename, "w");
			   if (!p_File) {
				fprintf(stderr,"  \nError: Could not open output file \"%s\", using default output (stdin)...\n",filename);
				p_File = stdout;
			   }
		  }
		  fprintf(p_File,"Species 1\tSpecies2\tWeight\n");
		  
		  if (strcmp(filename,"stdout") != 0)
			   fclose(p_File);
     }
     

     
     if (deg == 0) { //leaf node
	       return;
	  
	  } 
	  else if (deg == 1) { //undefined
	       fprintf(stderr,"Error: Node has only one child\n");
	       exit(EXIT_FAILURE);	  
	  } 
	  else if (deg == 2) { //binary case
			
	       if (F[idI][0] == NULL && F[idI][1] == NULL) { //all pairs of leaves that belong to the optimal path-system are located in the children of the node
			  //printf("Node %d with no pair, call recursively...\n",idI);
			  backtrace(p_node->p_children[0], NULL, F, F2, N, S, W, MWM, filename);
			  backtrace(p_node->p_children[1], NULL, F, F2, N, S, W, MWM, filename);
	       } 
		   else { //a particular pair (x,y) that belongs to the optimal path-system goes through the node
			 
		    Node *x = F[idI][0];
			Node *y = F[idI][1];
						 
			 //printf("Printing pair that belongs to the optimal path-system: \n");
			 if (strcmp(filename,"stdout") == 0) {
			      p_File = stdout;	       
			 } 
			 else {
			      p_File = fopen (filename, "a+");
			      if (!p_File) {
				   fprintf(stderr,"  \nError: Could not open output file \"%s\", using default output (stdin)...\n",filename);
				   p_File = stdout;
			      }
			 }
			 fprintf (p_File,"%s\t%s\t%lf\n", x->p_label, y->p_label, W[x->idL][y->idL]);

			 if (strcmp(filename,"stdout") != 0)
			      fclose(p_File);
			 
			 //Call other subtrees hanging off the path from (x,y) recusirvely			
			 backtracePathToLeaf (p_node, x, F, F2, N, S, W, MWM, filename);		 
			 backtracePathToLeaf (p_node, y, F, F2, N, S, W, MWM, filename);
			
	       }
 
	  } 
	  else if (deg > 2) { //multifurcation
 
	       //printf("Node with ID %d has more than two children, not implemented yet\n",p_node->idA); 
		   
		   int MWMIndex = getMWMIndex(p_node, p_nodeExcluded);

		   for (int i = 0; i < deg; i++) { //for all pairs of the MWM
			 
			 //printf("Loop through all pairs of the MWM: i = %d\n", i);
			 
			 //node not set da excluded
			 
			 if ((MWMIndex == 0 && MWM[idM][MWMIndex][i] == -1) || (i != (MWMIndex-1) && MWM[idM][MWMIndex][i] == -1) ) { //must not be -1 in this setting, either part of a pair (v,v') or (v,v*)
			   	//fprintf(stderr,"Error: MWM entry NULL (MWM[%d][%d][%d]) in backtracing for node with ID %d and i=%d\n", idM, MWMIndex, i, p_node->idA,i);
				//fprintf(stderr, "MWMe nrty: %d\n",MWM[idM][MWMIndex][i]);
				//exit(EXIT_FAILURE);
				continue;
			 }

			 if (MWM[idM][MWMIndex][i] == i) { // edges (v, v*)
			    //printf("Child %d paired with itself, backtrace further...\n", i);
				backtrace(p_node->p_children[i], NULL, F, F2, N, S, W, MWM, filename);		   
			 } 
			 else { // edges (v, v')
			   
			   //solution of Qu bla is needed
			   int j = MWM[idM][MWMIndex][i];
			   
			   if (i > j) { //Skip, we already examined this pair, due to the symmetry
				 continue;
			   }

			    if (strcmp(filename,"stdout") == 0) {
					p_File = stdout;	       
				} 
				else {
					p_File = fopen (filename, "a+");
					if (!p_File) {
					  fprintf(stderr,"  \nError: Could not open output file \"%s\", using default output (stdin)...\n",filename);
					  p_File = stdout;
					}
				}
				
				//printf ("Reading F2[%d][%d][%d][0/1]...\n", idM, i, j);
								
				Node *x = F2[idM][i][j][0];
				Node *y = F2[idM][i][j][1];
				
				if (x == NULL || y == NULL) {
					fprintf(stderr,"  \nError: F2 matrix entry F2[%d][%d][%d][0/1] is NULL...\n",idM, i, j);
					exit(EXIT_FAILURE);	
				}
				
				double weightPair = W[x->idL][y->idL];
				fprintf (p_File,"%s\t%s\t%lf\n", x->p_label, y->p_label, weightPair);

				if (strcmp(filename,"stdout") != 0)
					  fclose(p_File);
				
				//Call other subtrees hanging off the path from (x,y) recusirvely
				backtracePathToLeaf (p_node, F2[idM][i][j][0], F, F2, N, S, W, MWM, filename);		 
				backtracePathToLeaf (p_node, F2[idM][i][j][1], F, F2, N, S, W, MWM, filename);
				  
			 } // end edges (v, v')
 
		   }  //end for all pairs of the MWM
	  } // end multifurcation case
}

void backtracePathToLeaf (Node *startNode, Node *endNode, Node ***F, Node *****F2, Node ***N, double *S, double **W, int ***MWM, char *filename) {
  		
	//printf("1. Start examining node %d recursively...\n",idA);
	//printf("Leaf x: %d...\n",x->idA);
	int endNode_idL = endNode->idL;
	Node *curNode = N[startNode->idA][endNode_idL]; // child from p_node on the way to x
	//printf("Nachfolger von %d auf Weg zu %d ist %d\n",idA,x->idA,curNode->idA);
	int curNode_idA = curNode->idA;
	//path from startNode to endNode
	while (curNode_idA != endNode->idA) {
		  Node *p_nextNode = NULL;
		  Node *p_nodeExcludedNew = NULL;
		  if (curNode->deg == 2) {
			p_nextNode = getUquer(curNode, endNode, N);
			//printf("Determine u_bar from node %d on the way to %d...\n",curNode->idA,x->idA);
			//printf("Call node %d recurively...\n",p_nextNode->idA);
			backtrace(p_nextNode, p_nodeExcludedNew, F, F2, N, S, W, MWM, filename);
		  } 
		  else {
			p_nodeExcludedNew = N[curNode_idA][endNode_idL];
			backtrace(curNode, p_nodeExcludedNew, F, F2, N, S, W, MWM, filename); //multifurcating node is called again without one node
		  }
		  curNode = N[curNode_idA][endNode_idL];
		  curNode_idA = curNode->idA;
	}
}
