/**
 * main.c
 * Belongs to the TARGETING software
 * http://www.bioinf.uni-leipzig.de/Software/Targeting
 *
 * @author Christian Arnold
 * @email achristian@bioinf.uni-leipzig.de
 * @company Professur fuer Bioinformatik, University of Leipzig
 */
 
   
/*           TARGETING Copyright (C) 2010  Christian Arnold

* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.

* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.

* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>
* or the file "GNU Public License.txt" in the "src" folder.
*/

#include "main.h"
#include "manopt.h"
#include "misc.h"
#include "tree.h"
#include "parseData.h"
#include "MPP.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#define MAX_LENGTH_SP 60
#define VERBOSE 0


int main (int argc, char** argv) {
  
	  /*____________________________________________________		
		             Parse command line options
	    ____________________________________________________
	   */	  
     
	  manopt_optionset optset;
	  manopt_arg *unflagged;  
	  
	  char *treeFile   = "tree.tre";
	  char *scoresFile = "";
	  char *outputFile = "stdout";
	  int mode = 0;
	  int prec = 2;

	  manopt_initoptionset(	&optset, argv[0], NULL,
				   "\n\n This software calculates an optimal path-system for a given phylogenetic tree and a set of pairwise scores\n For more details, see the file Documentation.pdf in the main folder\n",
				   " This software is free software for non-commercial use and relased under the GNU Public License, (C) 2010 Bioinformatik Leipzig\n See the following publication for more details:\n   Arnold, C. and Stadler,PF. 2010. Polynomial algorithms for the Maximal Pairing Problem: efficient phylogenetic targeting on arbitrary trees. Algorithms for Molecular Biology. In review.\n", 
				   VERSION,
				   " Please report bugs to achristian@bioinf.uni-leipzig.de");
	  manopt_blockseparator(&optset, "INPUT/OUTPUT");
	  manopt(&optset, REQSTRINGOPT, 0, 't', "tree"  , "path/filename of Newick tree file", "<file>", NULL, &treeFile);
	  manopt(&optset, REQSTRINGOPT, 0, 's', "scores", "path/filename of file with pairwise scores", "<file>", NULL, &scoresFile); 
	  manopt(&optset, REQUINTOPT  , 0, 'm', "mode"  , "mode (0, 1, or 2)", "<integer>", NULL, &mode); 
	  manopt(&optset, REQSTRINGOPT, 0, 'o', "output", "path/filename of output file", "<file>", NULL, &outputFile);
	  manopt(&optset, REQUINTOPT  , 0, 'p', "prec",   "precision for the pairwise weights (2-10, see Documentation)", "<integer>", NULL, &prec);  
	  unflagged = manopt_getopts(&optset, argc, argv);
	  if(unflagged->noofvalues > 1) {
		    manopt_help(&optset, "Error: Unknown argument(s) provided\n");
	  }
	  
	  printf("\n");
	  
	  printf(" _______________________________________________________________________________\n\n");
	  printf(" TARGETING software, Author: Christian Arnold (achristian@bioinf.uni-leipzig.de)\n");
	  printf(" _______________________________________________________________________________\n\n");	  
	  printf(" Use the --help function to see all command line options\n\n");
	  
	  printf(" Parsing command line arguments...\n");
	  
	  //Check if mode has valid option
	  if(manopt_isset(&optset, 'm', "mode")){
	       if (mode != 0 && mode != 1 && mode != 2) {
		    char error[] = "\n\n---\nError with option -m: Value must be either 1, 2, or 3 (see Help)\n---\n\n";
		    manopt_help(&optset, error);
	       }
	  } else {
	       printf("  Argument -m not found, using default mode \"%d\"...\n",mode);
	  }
	  
	  if (manopt_isset(&optset, 'p', "prec")) {
		  if (!(prec >= 2 && prec <= 10)) {
		    char error[] = "\n\n---\nError with option -p: Value must be between 2 and 10 (see Help)\n---\n\n";
		    manopt_help(&optset, error);
	       }
		
	  } else {	  
	       printf("  Argument -p not found, using default precision of \"%d\" decimal places\n", prec);
	  } 
	  
	  prec = (int) pow ((double) 10, (double) prec);
 
	       
	  if(!manopt_isset(&optset, 't', "tree")){
	       printf("  Argument -t not found, using default tree file \"%s\"...\n", treeFile);
	  } 
	  
	  if(!manopt_isset(&optset, 's', "scores")){
	       printf("  Argument -s not found, using default pairwise scores of 1 for all pairs...\n");
	  } 
	  
	  if(!manopt_isset(&optset, 'o', "output")){
	       printf("  Argument -o not found, using default output file (stdout)...\n");
	  }
	  
	  if (strcmp(scoresFile,"") == 0 && mode == 0) { // default weights if no scores files is present
		mode =  1;
		prec = -1; //No approximation necessary if default scores are used
	  }
	       
	  manopt_destructoptionset(&optset);
	  manopt_destructarg(unflagged);	
	  free(unflagged);

	  /*____________________________________________________		
		                Parse Newick tree
	    ____________________________________________________
	   */	  
	  Node *p_root=NULL;	  
	  p_root = parseNewick(treeFile,p_root);
	  
	  /*____________________________________________________		
		       Initialize all variables that are needed
		        for computing the optimal path-system
	    ____________________________________________________
	   */	 
	 
	  int numberLeaves  = 0;
	  int numberInternalNodes = 0;
	  int *p_counterLeaves = &numberLeaves;
	  int *p_counterInternal = &numberInternalNodes;
	  Node **pp_leafList = getNextLeafRek(p_root,NULL,p_counterLeaves, p_counterInternal);
	  
	  if (pp_leafList == NULL) {
	       fprintf(stderr,"Error\n");
	       exit(EXIT_FAILURE);
	  }
	  
	  int numberNodes = numberInternalNodes + numberLeaves;
	  int numberMultifurcatingNodes = 0;
	  int *p_numberMultifurcatingNodes = &numberMultifurcatingNodes;
	  
	  //Traverse tree and determine number of multifurcating nodes
	  initIdM(p_root, p_numberMultifurcatingNodes);
	  
	  if (VERBOSE)
		printTree (p_root);
	   
	  printf("\n  Tree statistics:\n");
	  printf("   Number of total vertices: %d\n",numberNodes);	  
	  printf("   Number of interior vertices: %d\n",numberInternalNodes);	  
	  printf("   Number of taxa/leaves: %d\n",numberLeaves);
	  
	  int nrows = numberNodes;
	  int ncols = numberLeaves;
	  
	  Node ***N = allocateMemoryMatrixNode(nrows, ncols);
	  initMatrixNode (N, nrows, ncols, NULL);	  
	  fillNMatrix(N,numberLeaves,pp_leafList,p_root->idA);

	  char **arrSpecies;
	  arrSpecies = malloc (numberLeaves * sizeof(char*));
	  
	  getTaxaNames (arrSpecies, pp_leafList, numberLeaves);

	  int treeIsBinary = isBinary(p_root);
	  if (treeIsBinary) {
	     printf("   Tree is binary\n");  
	  } else {
	     printf("   Tree is not binary (%d multifurcation(s) found)\n", numberMultifurcatingNodes);  
	  }
	  	  
	  printf("\n Pairwise weights based on ");	  
	  switch (mode) {
	  
		  case 0:
			  printf("specified scores file\n");
			  break;
		  case 1 :
			  printf("default weights (maximizes number of pairs)\n");
			  break;
		  case 2 :
			  printf("graph-theoretic distance (maximizes number of edges)\n");
			  break;
	  }
	  
	  /*
	   * The F matrix stores information that is needed for backtracing for binary nodes
	   * For each interior node u, either the IDs of the best scoring pair is saved in F[u] or F[u] is empty
	   */
	  nrows = numberInternalNodes;
	  ncols = 2;
	  Node ***F = allocateMemoryMatrixNode(nrows,ncols);
	  initMatrixNode(F,nrows,ncols,NULL);
	   
	  /*
	   * The Q matrix stores solutions of partial problems for multifurcating nodes
	   * This matrix is not needed if the tree is binary
	   */
	  double **Q = NULL;
	  if (!treeIsBinary) {
	       nrows = numberNodes;
	       ncols = numberLeaves;
	       Q = allocateMemoryMatrixFloat(nrows,ncols);
	       initMatrixFloat(Q,nrows,ncols,-1);
	  }
	    
	  /*
	   * The R matrix stores solutions of partial problems
	   */
	  nrows = numberNodes;
	  ncols = numberLeaves;
	  //ncols = numberNodes;
	  double **R = allocateMemoryMatrixFloat(nrows,ncols);
	  initMatrixFloat(R,nrows,ncols,-1);

	  /*
	   * S matrix
	   * 1-dimensional array that stores the score of the optimal path-system for any interior node
	   */
	  double *S;
	  nrows = numberNodes;
	  S = malloc (nrows * sizeof(double));
	  initArrayFloat(S,nrows,-1);
	  
  
	  /*
	   * W matrix
	   * 2-dimensional array that stores the weights between any pair of species
	   * The scores are read from a scores file if user specified that with the -scores option
	   */
	  nrows = ncols = numberLeaves;
	  double **W = allocateMemoryMatrixFloat(nrows,ncols);
	  //2 char arrays that somehow reflect the taxa and the taxa2Int translation
	  initMatrixFloat(W,numberLeaves,numberLeaves,1);
	  
	  //Parse score file and update scores if file is present
	  
	  if (mode == 0) {
		int onlyIntegerWeights = parseScoreFile(scoresFile,W,arrSpecies,numberLeaves,pp_leafList);
		
		if (onlyIntegerWeights != 0) { //Only Integer weights are used, no approximation necessary
		  prec = -1;
		  //printf("Only integer weights are used, no approximation necessary...\n");
		}
	  
	  } else if (mode == 1) { //Default weights of 1: maximize number of pairs
		; //Nothing to do here
		
	  } else if (mode == 2) {  // Maximize number of used edges
		initScoresMode (p_root, W, pp_leafList, numberLeaves);
	  }
	
 
	  //Init S, R, and Q matrix: all leaf nodes get score 0 in each of these matrices
	  for (int i=0;i< numberLeaves;i++) {
	       int idA = pp_leafList[i]->idA;
	       int idL = pp_leafList[i]->idL;
	       S[idA] = 0;
	       R[idA][idL] = 0;
	       if (!treeIsBinary) {
		    Q[idA][idL] = 0;
	       }
	  }

	  if (VERBOSE)
		 printMatrixFloat(W,numberLeaves,numberLeaves);
	  
	  int ***MWM = NULL;
	  double ***Q2  = NULL;
	  Node *****F2  = NULL;

	  int counterIdM = 0;
	  int *p_counterIdM = &counterIdM;
		  
	  if (!treeIsBinary) {	  

		  //Dynamic memory allocation
		  MWM = malloc(numberMultifurcatingNodes * sizeof(int **));
		  Q2  = malloc(numberMultifurcatingNodes * sizeof(double **));
		  F2  = malloc(numberMultifurcatingNodes * sizeof(Node ****));
		  
		  if(MWM == NULL || Q2 == NULL || F2 == NULL) {
			fprintf(stderr, "Error: Could not allocate memory\n");
			exit(EXIT_FAILURE);
		  }

		  mallocMatricesMultiFurcating (p_root, Q2, F2, MWM, p_counterIdM);
		  counterIdM = 0;
		  initMatricesMultiFurcating   (p_root, Q2, F2, MWM, p_counterIdM);
	  }
	  
	  /*____________________________________________________		
			  Traverse tree and calculate the optimal
		          path-system for every vertice
	    ____________________________________________________
	   */	
	  
	  //Create array of pointers that represent a postorder tree traversal
	  Node **pp_nodesPostorder = NULL;
	  pp_nodesPostorder = malloc(numberNodes * sizeof(Node*));
	  int nodeCounter = 0;
	  int *p_nodeCounter = &nodeCounter;
	  
	  traverseTreePostorder (pp_nodesPostorder, p_root,p_nodeCounter);

	  //Calculate optimal path-system for every internal vertice
	  printf("\n Calculating an optimal path-system...\n\n");
	  for (int i=0; i< numberNodes; i++) {
	       
	        calcOptPathSystem(pp_nodesPostorder[i], F, F2, N, Q, Q2, R, S, W, MWM, prec);
	  }
	  
	  backtrace (p_root, NULL, F, F2, N, S, W, MWM, outputFile);
	  printf("\n Calculation of an optimal path-system finished successfully\n");
	  
	  printf("\n Score of optimal path-system: %f\n",S[p_root->idA]);
	  
	  if (strcmp(outputFile,"stdout") == 0) {
		printf("\n The selected pairs of the optimal path-system have been printed to \"%s\".\n",outputFile);
	  } 
	  else  {
		printf("\n You may view the selected pairs of the optimal path-system in the selected output file \"%s\".\n",outputFile);
	  }

	  printf("\n\n Exit\n\n");	  

	/*____________________________________________________		
	 *		      Free Memory
	 * ____________________________________________________
	 */

	freeMemoryMatrix((void **) F, numberInternalNodes);
	freeMemoryMatrix((void **) N, numberNodes);
	freeMemoryMatrix((void **) R, numberNodes);
	freeMemoryMatrix((void **) W, numberLeaves);
	free(S);
	
	if (!treeIsBinary) {
	  freeMemoryMatrix((void **) Q, numberNodes);
	  counterIdM = 0;
	  freeMatricesMultiFurcating (p_root, F2, Q2, MWM, p_counterIdM);
	  free(F2);
	  free(Q2);
	  free(MWM);
	}
	
	free(pp_leafList);
	free(pp_nodesPostorder);	
	free(arrSpecies);
	freeTree(p_root);
	
	return EXIT_SUCCESS;
}