/**
 * parseData.c
 * Belongs to the TARGETING software
 * http://www.bioinf.uni-leipzig.de/Software/Targeting
 *
 * @author Christian Arnold
 * @email achristian@bioinf.uni-leipzig.de
 * @company Professur fuer Bioinformatik, University of Leipzig
 */
 
   
/*           TARGETING Copyright (C) 2010  Christian Arnold

* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.

* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.

* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>
* or the file "GNU Public License.txt" in the "src" folder.
*/

#include "tree.h"
#include "parseData.h"
#include "misc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>


#define MAX_LENGTH_TAXA_STR 60
#define BUFSIZE_LINE 150
#define MAX_WEIGHT 10000

int parseScoreFile (char *filename, double **W, char **arrSpecies, int numberLeaves, Node **pp_leafList) {

	int onlyIntegerWeights = 1;
  
	unsigned int totalPairs = numberLeaves * (numberLeaves-1) /2;
	
	if (strcmp(filename,"") == 0) {
	     return -1;
	}
	
	printf("\n Parse score file \"%s\"...\n",filename);
	
	FILE *pFile;
	pFile = fopen (filename, "r");
	if (!pFile) {
	    fprintf(stderr,"  \nError: Could not open scores file %s, using default scores of 1 for all pairs...\n",filename);
	    return -1;
	  }

	  char sp1[60];
	  char sp2[60];
	  double score;
	  unsigned int line_counter = 0;
	  char buff[BUFSIZE_LINE];
	  
	  //Read line by line
	  while ( fgets( buff, sizeof (buff), pFile ) != NULL ) {

	       if ( sscanf( buff, "%s %s %lf", sp1, sp2, &score) == 3 ) {
		    
				line_counter++;
				int isError = 0;
				 
				if (line_counter > totalPairs) {
					fprintf(stderr, "Too many lines read\n");
					exit(EXIT_FAILURE);
				}

				if (!isValidSpName(sp1, arrSpecies, numberLeaves)) {
					printf("Error: %s is not a valid species\n",sp1);
					isError = 1;
				} 
				if (!isValidSpName(sp2, arrSpecies, numberLeaves)) {
					printf("Error: %s is not a valid species\n",sp2);
					isError = 1;
				} 
				if (score <=0 || score >= MAX_WEIGHT) {
					printf("Error: Score \"%f\" in line %d is not greater than 0 or bigger than the maximal allowed weight (%d)\n",score,line_counter,MAX_WEIGHT);
					isError = 1;
				}
				
				if (isError) {
					fprintf(stderr,"Error while parsing the file \"%s\" in line %d, exit...\n\n",filename,line_counter);
					exit(EXIT_FAILURE);
				
				} 
				else { //Fill W matrix
				
				 int idL_sp1 = -1, idL_sp2 = -1;
				 int foundSp1 = 0, foundSp2 = 0;
				 
				 for (int i = 0; i < numberLeaves; i++) {
					  if (strcmp(pp_leafList[i]->p_label, sp1) == 0) {
						idL_sp1 = pp_leafList[i]->idL;
						foundSp1 = 1;
					  } 
					  else if (strcmp(pp_leafList[i]->p_label, sp2) == 0) {
						idL_sp2 = pp_leafList[i]->idL;
						foundSp2 = 1;
					  }
					  
					  if (foundSp1 == 1 && foundSp2 == 1)
						break;
				 }
				 
				 if (! (score == ceil(score) && score == floor(score))) {
					onlyIntegerWeights = 0;
					printf( "\n  WARNING: You specified non-integer weights. The weights for the maximal weighted matching problems may be approximated\n  to a particular number of decimal places (see the -prec parameter). See the Documentation for more information!\n");
				 }
				 
				 //translate species names
				 W[idL_sp1][idL_sp2] = W[idL_sp2][idL_sp1] = score;
				 
				}
			 
	       } 
		   else {
			 fprintf(stderr, "Error: Line %s not valid\n",buff);
			 exit(EXIT_FAILURE);
	       }
	  }

	  fclose (pFile);
	  
	  return onlyIntegerWeights;		
}

char* getline(char *filename) {
	
	unsigned int pos = 0;
	int c;

	FILE *pFile;
	pFile = fopen (filename, "r");
	if (!pFile) {
	  fprintf(stderr,"  \nError: Could not open file %s, exit...\n\n",filename);
	  exit(EXIT_FAILURE);
	}
	
	//Determine how many characters the line has, avoids calling realloc
	unsigned int num_characters = 0;
	while(fgetc(pFile) != '\n' && !feof(pFile)) {
	  num_characters++;
	}
	num_characters++; /*leave room for '\0'*/
	//printf("Number of characters in file that were counted:%d\n",num_characters);	
	
	//Now we know how big the char array has to be
	char *line = malloc (num_characters * sizeof(char));

	
	//Reset file pointer to beginning of file
	rewind(pFile);
	
	//Read line
	while((c = fgetc(pFile)) != EOF) {
	    if(c == '\n')
		break;

	    if(pos < num_characters) {
		line[pos++] = c;
	    }
	}

	if(c == EOF && pos == 0){
	    fprintf(stderr,"\nError in getline: The file is empty, exit...\n\n");
	    exit(EXIT_FAILURE);
	}

	line[pos] = '\0';
	fclose(pFile);
	//printf("The first line from the file \'%s\' is %s\n",filename, line );
	
	return line;
}


Node * parseNewick(char *filename,Node *p_root) {
	
	const char *newickStr;
	
	if ((newickStr = getline(filename)) != '\0') {
	  //printf("Number of characters in file that were read:%d\n",strlen(newickStr));	
	  printf("\n A Newick tree has been found in the file \"%s\"...\n",filename);
	} 
	else {
	  fprintf(stderr," Error with reading the first line,exit\n");
	  exit(EXIT_FAILURE);
	}

	int idA = 0, idL = 0, idI = 0;
	int *p_idA = &idA; // Pointer to ID field that counts all vertices
	int *p_idL = &idL; // Pointer to ID field that counts only leaves
	int *p_idI = &idI; // Pointer to ID field that counts only interior vertices	
	
	checkNewickString(newickStr);
	
	p_root = malloc(sizeof (Node));
	p_root->idA = *p_idA;
	p_root->idL = -1;
	p_root->idI = *p_idI;
	p_root->p_parent = NULL;
	p_root->p_label = NULL;
	
	parseNewickRek(newickStr, 1,strlen(newickStr)-3,p_idA,p_idL,p_idI,p_root); //3 because 2x ( + 1x ;
	
	free((void *) newickStr);
	return p_root;
}


void parseNewickRek(const char newickString[], int i, int j, int *p_idA, int *p_idL, int *p_idI, Node *p_root)  {
	
	if (p_root == NULL) {
	  fprintf(stderr,"Error: Parent root is NULL\n");
	  exit(EXIT_FAILURE);
	} 
  
    int stringLength = strlen(newickString);
		
	//printf("Calling parseNewickRek with i=%d, j=%d, node_id=%d...\n",i,j,*p_idA);
	//printf("  %s, total length:%d\n",newickString,stringLength);
	
	char *subtree = substring(newickString, i,j-i+1);
	free(subtree);
	
	int deg = 0;	
	Node **p_children = NULL;
	
	while (i <= j) {
	  	
		//printf("  i=%d, Character at this pos.:%c\n",i,newickString[i]);
		//add node
		//printf("  Add node...\n");
		
		//add edge from p_root to this new child
		  Node *p_curNode = malloc(sizeof(Node));
		  //memcpy(p_curNode,&cur_node,sizeof(Node));  
		  p_curNode->p_parent = p_root;
		  p_curNode->idA = ++(*p_idA);

		int k=0;

		if (newickString[i] == '(') { //hit a subtree

			//printf("  Hit a subtree and search for corresponding closing bracket...\n");
			//printf("  Size of p_children: %d\n",sizeof(p_children));				
			//printf("  Increase dim children to %d for node id %d\n",(deg+1),node_id_orig);			
			deg++;
			
			//printf(" Reallocate more memory:from %d to %d...\n",sizeof(p_children),(deg *  sizeof(Node*)));
			p_children = realloc(p_children, deg * sizeof(Node*));
			//printf(" Memory after alloc: %d...\n",sizeof(p_children));
			
			//p_children[deg-1] = malloc(sizeof(Node*));
			p_children[deg-1] = p_curNode;
			
			int br_counter = 0;
			
			//set k to the position of the balancing closing-bracket for i
			for (int cur_pos = i; cur_pos < stringLength; cur_pos++) {
				
				if ( newickString[cur_pos] =='(') {
					br_counter++;
				} 
				else if ( newickString[cur_pos] ==')') {
					br_counter--;
				}
				if (br_counter == 0) {
					k = cur_pos;
					break;
				}					
				if (cur_pos == (stringLength-1)) { //-1 ??
					fprintf(stderr,"Error in Newick tree: Could not find closing bracket...\n");
					exit(EXIT_FAILURE);
				}
			}
			//printf("  Position k of closing bracket: %d\n",k);

			p_curNode->p_label = NULL;
			p_curNode->idL = -1;
			p_curNode->idI = ++(*p_idI);
			
			//return node*, how to use this pointer here?
			//printf("  Call parseNewickRek with i=%d, j=%d, node_id=%d\n",(i+1),(k-1),(*p_idA+1));
			parseNewickRek(newickString,i+1,k-1,p_idA,p_idL,p_idI,p_curNode);
			//printf("  Finished recursive call\n");
			
			// further shift k to the right until the end of the (optional) branch length definition (e.g., "):1.345")			
			
			if (newickString[k+1] == ':') {
			  //printf("  Search for rbanch length ende...\n");
			  int k_new = j;
			  for (int cur_pos = k; cur_pos <= j; cur_pos++) {
				  if ( newickString[cur_pos] == ',') {
					  k_new = cur_pos-1;
					  break;
				  } 
			  }
			  //Update k: It now points either to the end of the string or to the next ","
			  //printf("  Set k from %d to %d...\n",k,k_new);
			  k = k_new;
			  //printf("  Further shift k to the end of the branch length def...new position:%d\n",k);
			} 
			else if (newickString[k+1] != ')' && newickString[k+1] != ','){
			     printf("Error: Internal branch label found, this is not supported, exit...\n");
			     exit(EXIT_FAILURE);			     
			}

		} 
		else { //hit a leaf
		
			//printf("  Hit a leaf, increase dim children to %d for node id %d\n",(deg+1),node_id_orig);
			deg++;
			
			//printf("  Size of p_children: %d\n",sizeof(p_children));
		      
			//printf(" Reallocate more memory:from %d to %d...\n",sizeof(p_children),(deg *  sizeof(Node*)));
			p_children = realloc(p_children, deg * sizeof(Node*));

			p_children[deg-1] = p_curNode;
		  
			//printf("  Hit a leaf...\n");
			
			p_curNode->deg = 0;
			p_curNode->idL = (*p_idL)++;
			p_curNode->idI = -1;			
			k=j;
			
			//printf("  Set k to %d...\n",k);
			
			int branchLengthDefined = 0;
						
			//Find next "," or ":" position
			for (int cur_pos = i; cur_pos <= j; cur_pos++) {
				if ( newickString[cur_pos] == ',' || newickString[cur_pos] == ':') {
					k = cur_pos-1;
					//printf("  Found , or :  at pos. %d\n",cur_pos);
					if (newickString[cur_pos] == ':') {
					  //printf("  Branch length defined...\n");
					  branchLengthDefined = 1;
					}
					break;
				}	
			}
			
			if (k<i) {
			  printf("  Error: k < i: %d < %d\n",k,i);
			  exit(EXIT_FAILURE);
			}

			char *nodeLabelFull;
			nodeLabelFull = substring(newickString, i,k-i+1);
			//printf("  Node label:%s\n",trimWhitespace(nodeLabelFull));
			p_curNode->p_label = trimWhitespace(nodeLabelFull);
			
			//printf("Node labekl in element: %s\n",p_curNode->p_label);
			
			int found_end = 0;
			
			if (branchLengthDefined == 1) {
			    //printf(" Search for end of branch length def...Range: %d - %d\n",(k+1),j);
			    //go to pos. after branch length def.
			    for (int cur_pos = k+1; cur_pos <= j; cur_pos++) {
				  if ( newickString[cur_pos] == ',' || newickString[cur_pos] == ')') {
					    //printf("  Found branch length def. between pos. %d and %d\n",(k+2),(cur_pos-1));
					    i= cur_pos-1;
					    k=i;
					   // printf("  Set i to %d after end of branch length has been found...\n",i);
					    found_end = 1;
					    break;
				  }
			    }
			    if (found_end == 0) {
			      k=j;
			    }
			}	
		}
		
		//printf("  Increment i to %d, i orig:%d\n",(k+1),i_orig);

		i = k+1; //set i to the next "," position
		if (i<j) {
			i++; //Increment i
		}
		
	}
	//printf("%d children for node with id %d\n",deg,node_id_orig);
	p_root->deg = deg;
	p_root->p_children = p_children;
		
}


void checkNewickString (const char *newickStr) {
	
	int is_error = 0;
	int bracket_counter = 0;
	
	int strLen = strlen(newickStr);
	
	printf("  Check Newick string...");
	
	//Test 1
	if (newickStr[strLen-1] != ';' || newickStr[strLen-2] == ';') {
		fprintf(stderr, "  \nError with Newick string: The last character must be \";\"!\n");
		is_error = 1;
	}

	//Test 2
	for (int cur_pos = 0; cur_pos < strLen; cur_pos++) {
				
		if (newickStr[cur_pos] == ' ' && is_error == 0) {
			fprintf(stderr, "  \nError with Newick string: Whitespace character detected, whitespaces are not allowed!\n");
			is_error = 1;
		}
	  
	    if ( newickStr[cur_pos] =='(') {
			bracket_counter++;
		} else if ( newickStr[cur_pos] ==')') {
			bracket_counter--;
		}
		if (bracket_counter<0) {
			fprintf(stderr, "  \nError with Newick string: Too much closing brackets!\n");
			is_error = 1;
		}
	}
	
	if (bracket_counter != 0) {
		fprintf(stderr, "  \nError with Newick string: Different number of opening and closing brackets!\n");
		is_error = 1;
	}
	
	//Test 3
	//$reg = "/e/i"; #e is allowed, since 4.5e-06 can be a branch length
	//if (preg_match($reg,$tree,$parts)) {
	//	$reg2 = "/e-\d+/i"; #check if e does not only occur in the scientific number format
	//	if (!preg_match($reg2,$tree,$parts2)) {
	//		return "Tree contains a letter, branch labels are now allowed";
	//	}
	//}


	//Test 4
	//check if all paranthesis have the right position
	//$reg = "/[^,\(]\(/"; // [nicht , oder ( ] (
	//if (preg_match($reg,$tree,$parts)) {
	//	return "At least one paranthesis has a wrong position!";
	// }
	
	//Test 5
	for (int cur_pos = 0; cur_pos < strLen-1; cur_pos++) {
				
		if ( newickStr[cur_pos] =='(' && newickStr[cur_pos+1] ==')') {
			fprintf(stderr, "  \nError with Newick string: Empty pair of \"(\" and \")\" found!\n");
			is_error = 1;
		} 
	}
		
	//Summary
	if (is_error) {
		fprintf(stderr, "  \nThe Newick string is not valid, exit...\n");
		exit(EXIT_FAILURE);
	} else {
		printf("OK\n");
	}

}