/**
 * tree.c
 * Belongs to the TARGETING software
 * http://www.bioinf.uni-leipzig.de/Software/Targeting
 *
 * @author Christian Arnold
 * @email achristian@bioinf.uni-leipzig.de
 * @company Professur fuer Bioinformatik, University of Leipzig
 */
 
   
/*           TARGETING Copyright (C) 2010  Christian Arnold

* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.

* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.

* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>
* or the file "GNU Public License.txt" in the "src" folder.
*/

#define MAX_LENGTH_TAXA_STR 100

#include "tree.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

void printNode (Node *node,int spaces) {	

	int i=0;
	while (i++<spaces)
		printf(" ");

	if (node==NULL) {
		fprintf(stderr,"Error: Node is NULL\n");
		exit(EXIT_FAILURE);
	}
	 
	int id_parent= node->p_parent==NULL? -1: node->p_parent->idA;   
	char *label = NULL;

	if (node->p_label != NULL) {
		label = node->p_label;
	} 
	else {
		label="(none)";
	}
      
      printf("Print node...IdA:%d IdL:%d IdI:%d IdM: %d Dim:%d Parent:%d Label:%s\n",node->idA,node->idL,node->idI,node->idM,node->deg,id_parent,label);
}

Node** getNextLeafRek(Node *p_node, Node **pp_leafList,int *counterLeaves, int *counterInternal) {
  
  if (p_node->deg == 0) {  
     (*counterLeaves)++;
     pp_leafList = realloc(pp_leafList,(*counterLeaves) * sizeof(Node*));
    if (!pp_leafList) {
	  fprintf(stderr,"Error with reallocating\n");
	  exit(EXIT_FAILURE);
    }
     pp_leafList[(*counterLeaves)-1] = p_node;

  } else {
    if (counterInternal != NULL)
		(*counterInternal)++;
    for (int i=0;i<p_node->deg;i++) {
      pp_leafList = getNextLeafRek(p_node->p_children[i],pp_leafList,counterLeaves,counterInternal);
    }
  }
  return pp_leafList;
 
}

void getTaxaNames (char **taxaNames, Node** pp_leafList, int numberTaxa) {
        
     for (int i=0; i< numberTaxa; i++) {
	  taxaNames[i] = pp_leafList[i]->p_label;
     }
}

Node *getUquer (Node *p_internal, Node *p_leaf, Node ***N) {
     
     Node *u = NULL;
     
     if (p_internal == NULL || p_leaf == NULL || p_internal->deg > 2 || p_leaf->deg != 0) {
	  fprintf(stderr,"Error in getUquer %d\n",p_internal->deg);
	  exit(EXIT_FAILURE);	  	  
     }
     
     Node *childToLeaf = N[p_internal->idA][p_leaf->idL];
     for (int i=0; i< p_internal->deg;i++) {
	if (p_internal->p_children[i] != childToLeaf) {
	       u = p_internal->p_children[i]; 
	}
     }
     
    if (u == NULL) {
	  fprintf(stderr,"Error in getUquer: u null\n");
	  exit(EXIT_FAILURE);	  	  
     } 
     
     return u;
}


void printTree (Node *p_root) { 
	
	printf("\nPrint whole tree...\n");
	if (p_root==NULL) {
	  fprintf(stderr,"Error: Root is NULL\n");
	  exit(EXIT_FAILURE);
	}
	printTreeRek(p_root,1);
	printf("Done\n");
}

void printTreeRek (Node *p_root, int spaces) { 
		
	printNode (p_root,spaces);
	if (p_root->deg > 0) {		
		spaces++;
		for (int i=0;i<p_root->deg;i++) {				
			printTreeRek(p_root->p_children[i], spaces);
		}
	}
}

void freeTree (Node *p_node) {
  
  for (int i=0; i<p_node->deg;i++) {
	  freeTree (p_node->p_children[i]);
	  if (i==(p_node->deg-1)) {
	      free(p_node->p_children);
	  }
  }
  if (p_node->p_label !=NULL) {
     free(p_node->p_label);  
  }
  free(p_node);

}

int isBinary(Node *p_node) {
     
	if (p_node->deg > 2) {
		return 0;
	} 
	else {
		for (int i=0; i< p_node->deg;i++) {
			int k = isBinary(p_node->p_children[i]);
			if (k == 0) {
				return 0;
			}       
		  }
     }
     
     return 1;    
}
  
void traverseTreePostorder (Node **p_nodes, Node *p_node, int *counter) {
     
     int deg = p_node->deg;
	 
	 // Postorder tree traversal
	 // 1. Traverse all subtrees from left to right
	 // 2. Visit the root  
     
	if (deg == 0) {
		p_nodes[(*counter)++] = p_node;
	} 
	else {
		for (int i=0;i<deg;i++) {
		   traverseTreePostorder (p_nodes, p_node->p_children[i], counter);
		}
		p_nodes[(*counter)++] = p_node;
	}
}

Node ** getLeavesInSubtree (Node *p_node, Node **pp_leafList, int *p_counterInternal) {
      
	if (p_node->deg == 0) {
		(*p_counterInternal)++;
		pp_leafList = realloc(pp_leafList, (*p_counterInternal) * sizeof(Node*));
		if (!pp_leafList) {
			fprintf(stderr,"Error with reallocating\n");
			exit(EXIT_FAILURE);
		}
		pp_leafList[(*p_counterInternal-1)] = p_node;
	} 
	else {
		for (int i = 0; i < p_node->deg; i++) {
		   pp_leafList = getLeavesInSubtree (p_node->p_children[i], pp_leafList, p_counterInternal);
		}
	}
	 
     return pp_leafList;
     
}

void initIdM(Node *p_node, int *p_counterIdM) {
  
  if (p_node->deg == 0) {
	p_node->idM = -1;
	return; 
  } 
  else {
  
	if (p_node->deg == 2) {
		p_node->idM = -1;
	}
	else { //deg > 2, multifurcating
		p_node->idM = (*p_counterIdM)++;
	}
  
	for (int i = 0; i < p_node->deg; i++) {
      initIdM(p_node->p_children[i], p_counterIdM);
    }
	
  }
 
}

void mallocMatricesMultiFurcating (Node *p_node, double ***Q2, Node *****F2, int ***MWM, int *counter_idM) {

  if (p_node->deg < 3) {
		
		for (int i = 0; i < p_node->deg; i++) {
			mallocMatricesMultiFurcating (p_node->p_children[i], Q2, F2, MWM, counter_idM);		
		}
	}
	else { //Multifurcation
		
		// 1. MWM matrix		
		int dim2 = p_node->deg + 1; //one for S_u, deg many for Q_ux
		int dim3 = p_node->deg; //matching partner of each child v of u 		
		int index = (*counter_idM);

		MWM[index] = malloc(dim2 * sizeof(int **));
		if(!MWM[index]){
			fprintf(stderr, "Error: Could not allocate memory, exit...\n");
			exit(EXIT_FAILURE);
		}
		for(int j = 0; j < dim2; j++) {
			MWM[index][j] = malloc(dim3 * sizeof(int*));
			if(!MWM[index][j]){
				fprintf(stderr, "Error: Could not allocate memory, exit...\n");
				exit(EXIT_FAILURE);
			}
		}
		
		// 2. Q2 and F2 matrix
		dim2 = dim3 = p_node->deg;
		Q2[index] = malloc(dim2 * sizeof(double *));
		F2[index] = malloc(dim2 * sizeof(Node ***));
		if(! (Q2[index] || F2[index])) {
			fprintf(stderr, "Error: Could not allocate memory, exit...\n");
			exit(EXIT_FAILURE);
		}
		for(int j = 0; j < dim2; j++) {
			Q2[index][j] = malloc(dim3 * sizeof(double));
			F2[index][j] = malloc(dim3 * sizeof(Node **));			
			if(! (Q2[index][j] || F2[index][j])) {
				fprintf(stderr, "Error: Could not allocate memory, exit...\n");
				exit(EXIT_FAILURE);
			}
			
			for(int k = 0; k < dim3; k++) {
				F2[index][j][k] = malloc(2 * sizeof(Node *));			
				if(!F2[index][j][k]) {
					fprintf(stderr, "Error: Could not allocate memory, exit...\n");
					exit(EXIT_FAILURE);
				}
			}
		}
	
		(*counter_idM)++;
		
		for (int i = 0; i < p_node->deg; i++) {
			mallocMatricesMultiFurcating (p_node->p_children[i], Q2, F2, MWM, counter_idM);		
		}
  }

}

void initMatricesMultiFurcating (Node *p_node, double ***Q2, Node *****F2, int ***MWM, int *counter_idM) {

   if (p_node == NULL) {
	  fprintf(stderr, "Error in initMatricesMultiFurcating: Node is NULL\n");
	  exit(EXIT_FAILURE);
   }
   
   if (p_node->deg < 3) {
		
		for (int i = 0; i < p_node->deg; i++) {
		     //printf("  initMatricesMultiFurcating: Calling child node with ID %d (idM:%d)\n", p_node->p_children[i]->idA, p_node->p_children[i]e->idM);
			initMatricesMultiFurcating (p_node->p_children[i], Q2, F2, MWM, counter_idM);		
		}
	}
	else { //Multifurcation
		
		// 1. MWM matrix		
		int dim2 = p_node->deg + 1; //one for S_u, deg many for Q_ux
		int dim3 = p_node->deg; //maximum number of pairs in the matching 
		int dim4 = 2;
		
		int index = (*counter_idM);

		for (int j = 0; j < dim2; j++) {
			for (int k = 0; k < dim3; k++) {			  			    
			  MWM[index][j][k] = -1;
			}
		}
		
		// 2. Q2 and F2 matrix
		dim2 = dim3 = p_node->deg;

		for(int j = 0; j < dim2; j++) {
			
			for(int k = 0; k < dim3; k++) {
				
			  Q2[index][j][k]= -1;	
			  
			  for(int l = 0; l < dim4; l++) {
					F2[index][j][k][l]= NULL;								
				}
			  						
			}
		}
		
		(*counter_idM)++;
		
		for (int i = 0; i < p_node->deg; i++) {
			initMatricesMultiFurcating (p_node->p_children[i], Q2, F2, MWM, counter_idM);		
		}
		
  }

}

void freeMatricesMultiFurcating (Node *p_node, Node *****F2, double ***Q2, int ***MWM, int *counter_idM) {
  
  if (p_node->deg < 3) {
		
		for (int i = 0; i < p_node->deg; i++) {
			freeMatricesMultiFurcating (p_node->p_children[i], F2, Q2, MWM, counter_idM);		
		}
	}
	else { //Multifurcation

		// 1. MWM matrix		
		int dim2 = p_node->deg + 1; //one for S_u, deg many for Q_ux
		int dim3;
		int index = (*counter_idM);

		for(int j = 0; j < dim2; j++) {		
			free(MWM[index][j]);
		}
		
		free(MWM[index]);
		
		// 2. Q2 and F2 matrix
		dim2 = dim3 = p_node->deg;

		for(int j = 0; j < dim2; j++) {
			free(Q2[index][j]);
			
			for(int k = 0; k < dim3; k++) {
			  free(F2[index][j][k]);
			}
			free(F2[index][j]);
		}	
		
		(*counter_idM)++;	

		free(Q2[index]);
		free(F2[index]);
		
		for (int i = 0; i < p_node->deg; i++) {
			freeMatricesMultiFurcating (p_node->p_children[i], F2, Q2, MWM, counter_idM);			
		}
  }

}

void initScoresMode (Node *p_root, double **W, Node **pp_leafList, int numberLeaves) {
  
  //Get maximum tree depth
  int treeDepth = getTreeDepth (p_root, pp_leafList, numberLeaves) + 1; //leaf ID is also saved, so we need one element extra
  //printf("Tree depth: %d\n", treeDepth);
  int arr_pathToRoot [numberLeaves] [treeDepth];
  
  int counter = 0;
  
  //create array with the paths from each leaf node to the root
  for (int i = 0; i < numberLeaves; i++) {
	
	  for (int j = 0; j < treeDepth; j++) { //init array
		arr_pathToRoot [i] [j] = -1;
	  }
	
	  Node *curNode = pp_leafList[i];
	  counter = 0;
	  
	  while (curNode != p_root) {		  		  	  
		  arr_pathToRoot[i][counter] = curNode->idA;
		  curNode = curNode->p_parent;
		  counter++;
		  
		  if (curNode == p_root) {
			arr_pathToRoot[i][counter] = p_root->idA;
		  }
	  }
	  
		// 	  printf("Leaf %s: ",pp_leafList[i]->p_label);
		// 	  for (int j = 0; j < treeDepth; j++) { //init array
		// 		printf(" %d", arr_pathToRoot [i] [j]);
		// 	  }
		// 	  printf("\n");
	
  }
  
  //compare arrays and calculate distance for each pair of leaves
  for (int i = 0; i < numberLeaves; i++) { //leaf i
	
	  for (int j = i + 1; j < numberLeaves; j++) { //leaf j
		
			int dist = 0;
			int foundAnc = 0;
		  
			for (int k = 0; k < treeDepth; k++) { //path from leaf i to root
			  
			  if (foundAnc == 1 || arr_pathToRoot[i][k] == -1) {
				continue;
			  }
			  
			  for (int l = 0; l < treeDepth; l++) { //path from leaf j to root
					  
				if (arr_pathToRoot[j][l] == -1) {
				  break;
				}
				
				if (arr_pathToRoot[i][k] == arr_pathToRoot[j][l]) {
				  foundAnc = 1;
				  dist = k + l;
				  //printf("  Found ancestor node with ID %d:  dist=%d\n", arr_pathToRoot[i][k], dist);
				  break;
				}
					 
			  }
			  
			}
			
			if (foundAnc == 0) {
			  	fprintf(stderr, "Error: Ancestor not found for pair of species, exit...\n");
				exit(EXIT_FAILURE);
			}
		  	  
		  Node *p_i = pp_leafList[i], *p_j = pp_leafList[j];
		
		  W [p_i->idL] [p_j->idL] = W [p_j->idL] [p_i->idL] = dist;
	
	  }
  }
  
}

int getTreeDepth (Node *p_root, Node **pp_leafList, int numberLeaves) {
  
    int curDepth = 0, maxDepth = 0;
  
	for (int i = 0; i < numberLeaves; i++) {
	  
		Node *curNode = pp_leafList[i];
		
		curDepth = 0;
		
		while (curNode != p_root) {
			curNode = curNode->p_parent;
			curDepth++;
		}
		
		if (curDepth > maxDepth) {
		  maxDepth = curDepth;
		  
		}	  
	}
	
	return maxDepth;
	
}